import os
from typing import Optional
from zoneinfo import ZoneInfo

import httpx
import yaml
from dateutil import parser
from dotenv import find_dotenv, load_dotenv

load_dotenv(find_dotenv())

OWNER = os.getenv("OWNER", None)
TOKEN = os.getenv("EXPIRING_TOKEN", None)
REPO = os.getenv("REPOSITORY", None)
BASE = "https://api.github.com/repos"
URL = f"{BASE}/{OWNER}/{REPO}/contents/justices/list.yaml"
HEADERS = {
    "Authorization": f"token {TOKEN}",  # bearer
    "Accept": "application/vnd.github.VERSION.raw",  # raw file
}


def get_justice_list_url() -> Optional[str]:
    """Gets the proper URL if environment variables are set

    Returns:
        Optional[str]: URL string

    >>> from lawsql_cases_justices import get_justice_list_url
    >>> assert URL == get_justice_list_url()
    True
    """
    return URL if all([OWNER, TOKEN, REPO]) else None


def get_justice_api_response() -> Optional[httpx.Response]:
    """Assumes the existence of a valid URL to secure the response object.

    Returns:
        Optional[httpx.Response]: The response object consisting of byte data from the yaml file online

    >>> from lawsql_cases_justices import get_justice_api_response
    >>> r = get_justice_api_response()
    >>> r
    <Response [200 OK]>
    """
    if not (url := get_justice_list_url()):
        return None
    with httpx.Client() as client:
        return client.get(url, headers=HEADERS)


def get_justice_data() -> Optional[dict]:
    """With properly set .env values, get the updated list of justices.

    Returns:
        Optional[dict]: [description]

    >>> from lawsql_cases_justices import get_justice_data
    >>> data = get_justice_data()
    >>> assert isinstance(data, dict)
    True
    >>> assert data.keys() == dict_keys(['last_modified', 'justices'])
    True
    >>> import datetime
    >>> assert isinstance(data['last_modified'], datetime.datetime)
    True
    >>> assert isinstance(data['justices'], list)
    True
    """
    from .config import get_justice_from_dict

    r = get_justice_api_response()
    if not r:
        return None

    return {
        "last_modified": parser.parse(r.headers["last-modified"]).astimezone(
            ZoneInfo("Asia/Manila")
        ),
        "justices": [
            get_justice_from_dict(i)
            for i in yaml.load(r.content, Loader=yaml.FullLoader)
        ],
    }



import datetime as dt

from dateutil.parser import parse


def parse_date_if_exists(text: str = None) -> dt.date | None:
    """If the variable contains text with more than 5 characters, parse possible date."""
    if not text:
        return None
    elif text and len(text) < 5:
        return None
    elif not (parsed := parse(text)):
        return None
    return parsed.date()



from .__main__ import (
    get_active_justices_with_date_name,
    get_id,
    get_justice_data_from_id,
    get_justice_id_date_name,
    get_justice_label_from_id,
    get_justice_surname_from_id,
    get_justices_on_date,
    get_justices_with_name,
)
from .api import (
    get_justice_api_response,
    get_justice_data,
    get_justice_list_url,
)
from .config import justices_tbl, load_justices
from .surname import get_surname



from datetime import date
from typing import Iterator, Optional

from sqlite_utils.db import NotFoundError

from .config import JUSTICES_TABLE
from .typos import is_typo


def get_id(d: date, x: str) -> Optional[int]:
    """With a "cleaned" surname, including considered typos, get primary key (pk) of matching Justice within the date provided."""
    return pk if (pk := is_typo(x) or get_justice_id_date_name(d, x)) else None


def get_justice_id_date_name(target_date: date, surname: str) -> Optional[int]:
    """With a justice's "cleaned" surname (s) and a valid date, if only one justice row is found, return primary key (pk) of that row"""
    l = list(get_active_justices_with_date_name(target_date, surname))
    return l[0]["pk"] if l and len(l) == 1 else None


def get_justices_with_name(last_name: str) -> Iterator[dict]:
    return JUSTICES_TABLE.rows_where(
        """--sql
        lower(last_name) = lower(:last_name_param) or
        lower(replace(last_name, '-', ' ')) = lower(:last_name_param);
        """,
        {"last_name_param": last_name},
        select="pk, last_name, suffix",
    )


def get_justices_on_date(target_date: date) -> Iterator[dict]:
    return JUSTICES_TABLE.rows_where(
        """--sql
        date(start_term) < date(:target_date_param) and
        date(:target_date_param) < date(coalesce(end_term, retire_date));
        """,
        {"target_date_param": target_date},
        select="pk, last_name, suffix",
    )


def get_active_justices_with_date_name(
    target_date: date, last_name: str
) -> Iterator[dict]:
    return JUSTICES_TABLE.rows_where(
        """--sql
        (
            date(start_term) < date(:target_date_param) and
            date(:target_date_param) < date(coalesce(end_term, retire_date))
        ) and
        (
            lower(last_name) = lower(:last_name_param) or
            lower(replace(last_name, '-', ' ')) = lower(:last_name_param)
        );
        """,
        {"target_date_param": target_date, "last_name_param": last_name},
        select="pk, last_name, suffix",
    )


def get_justice_data_from_id(pk: int) -> Optional[dict]:
    """Get Justice profile row from primary key (pk), if it exists."""
    try:
        return JUSTICES_TABLE.get(pk)
    except NotFoundError:
        return None


def get_justice_surname_from_id(pk: int) -> Optional[str]:
    """Get Justice surname from primary key (pk), if it exists."""
    return row["last_name"] if (row := get_justice_data_from_id(pk)) else None


def get_justice_label_from_id(pk: int) -> Optional[str]:
    """Get Justice surname from primary key (pk), if it exists."""
    if row := get_justice_data_from_id(pk):
        if row["suffix"]:
            return f"{row['last_name']}, {row['suffix']}"
        return row["last_name"]
    return None



import os
from pathlib import Path
from typing import Iterator

import yaml
from dateutil.relativedelta import relativedelta as rd
from dotenv import find_dotenv, load_dotenv
from sqlite_utils import Database
from sqlite_utils.db import Table

from .utils import parse_date_if_exists

load_dotenv(find_dotenv())

SC_JUSTICE_LIST = Path().home().joinpath(os.getenv("SC_JUSTICES"))
db_path = Path().home().joinpath(os.getenv("PATH_TO_DB"))
DB = Database(db_path, use_counts_table=True)


MAX_AGE = 70
MAX_LABEL_LENGTH = 50


def justices_tbl():
    tbl = DB["Justices"]
    if tbl.exists():
        return tbl

    tbl.create(
        columns={
            "pk": int,
            "first_name": str,
            "last_name": str,
            "full_name": str,
            "suffix": str,
            "gender": str,
            "start_term": str,
            "end_term": str,
            "chief_date": str,
            "birth_date": str,
            "retire_date": str,
        },
        pk="pk",
    )

    for i in [
        "pk",
        "last_name",
        "gender",
        "start_term",
        "end_term",
        "retire_date",
    ]:
        tbl.create_index([i])

    if not tbl.has_counts_triggers:
        tbl.enable_counts()

    return load_justices(tbl)


def get_justice_records_from(p: Path) -> Iterator[dict]:
    "Remove unessential contents from .yaml file path, e.g. 'Justice' and 'Replacing'"
    text = p.read_text()
    records = yaml.load(text, Loader=yaml.SafeLoader)
    for record in records:
        yield get_justice_from_dict(record)


def get_justice_from_dict(i: dict) -> dict:
    i.pop("Justice")
    i.pop("Replacing")
    return {
        "pk": int(i.pop("#")),
        "first_name": i.pop("First Name").strip(),
        "last_name": i.pop("Last Name").strip(),
        "suffix": s.strip() if (s := i.pop("Suffix")) else None,
        "gender": i.pop("Gender").strip(),
        "start_term": parse_date_if_exists(i.pop("Start of term")),
        "end_term": (_end := parse_date_if_exists(i.pop("End of term"))),
        "chief_date": parse_date_if_exists(i.pop("Appointed chief")),
        "birth_date": (dob := parse_date_if_exists(i.pop("Born"))),
        "retire_date": dob + rd(years=MAX_AGE) if dob else _end,
    }


def load_justices(tbl: Table):
    return tbl.insert_all(
        get_justice_records_from(SC_JUSTICE_LIST),
        pk="pk",
        not_null=("first_name, last_name, start_term"),
        replace=True,
    )


JUSTICES_TABLE = justices_tbl()



from typing import Optional

from .patterns import IS_JBL


def is_typo(text: str) -> Optional[int]:
    """Since some surnames are not properly fetched, are nuanced, or have bad spelling, get the primary key on match of the given text"""
    return possible_pattern(text) or bad_spelling(text)


def bad_spelling(text: str):
    if text == "Ynares-Satiago":
        return 145  # Ynares-Santiago
    elif text == "Bellosiillo":
        return 127  # Bellosillo
    elif text == "Ferran":
        return 111  # Fernan
    elif text == "John":
        return 23  # Fernan
    elif text == "Avancea'A":
        return 18
    return None


def possible_pattern(text: str) -> Optional[int]:
    if IS_JBL.search(text):
        return 64
    elif "Bengzon" in text.title():
        return 76
    elif "Bautista" in text.title():
        return 59
    elif "Reyes, A." in text.title():
        return 177
    elif "A. Reyes" in text.title():
        return 177
    elif "Reyes, R.T." in text.title():
        return 159
    return None
